@@ -1,3 +1,4 @@
+use std::collections::HashSet;
 use std::io::Write;
 
 use bstr::ByteSlice;
@@ -16,6 +17,7 @@ struct State {
     pull: bool,
     push: bool,
     dry_run: bool,
+    backup_capacity: Option<usize>,
 
     show_format: git_stack::config::Format,
     show_stacked: bool,
@@ -42,6 +44,7 @@ impl State {
         )
         .with_code(proc_exit::Code::CONFIG_ERR)?;
         let dry_run = args.dry_run;
+        let backup_capacity = repo_config.capacity();
 
         let show_format = repo_config.show_format();
         let show_stacked = repo_config.show_stacked();
@@ -153,6 +156,8 @@ impl State {
             pull,
             push,
             dry_run,
+            backup_capacity,
+
             show_format,
             show_stacked,
         })
@@ -202,12 +207,19 @@ pub fn stack(args: &crate::args::Args, colored_stdout: bool) -> proc_exit::ExitR
             return Err(proc_exit::Code::USAGE_ERR.with_message("Working tree is dirty, aborting"));
         }
 
-        let mut pulled = false;
+        let mut pulled_ids = HashSet::new();
         for stack in state.stacks.iter() {
+            let mut stack_pulled_ids = HashSet::new();
             if state.protected_branches.contains_oid(stack.onto.id) {
                 match git_pull(&mut state.repo, stack.onto.name.as_str(), state.dry_run) {
-                    Ok(_) => {
-                        pulled = true;
+                    Ok(pull_range) => {
+                        stack_pulled_ids.extend(
+                            state
+                                .repo
+                                .commits_from(pull_range.1)
+                                .take_while(|c| c.id != pull_range.0)
+                                .map(|c| c.id),
+                        );
                     }
                     Err(err) => {
                         log::warn!("Skipping pull of `{}`, {}", stack.onto.name, err);
@@ -219,25 +231,47 @@ pub fn stack(args: &crate::args::Args, colored_stdout: bool) -> proc_exit::ExitR
                     stack.onto.name
                 );
             }
+            if !stack_pulled_ids.is_empty() {
+                match drop_branches(
+                    &mut state.repo,
+                    stack_pulled_ids.difference(&pulled_ids).cloned(),
+                    &stack.onto.name,
+                    &state.branches,
+                    &state.protected_branches,
+                    state.dry_run,
+                ) {
+                    Ok(()) => {}
+                    Err(err) => {
+                        log::warn!("Could not remove branches obsoleted by pull: {}", err);
+                    }
+                }
+                pulled_ids.extend(stack_pulled_ids);
+            }
         }
-        if pulled {
+        if !pulled_ids.is_empty() {
             state.update().with_code(proc_exit::Code::FAILURE)?;
         }
     }
 
+    const BACKUP_NAME: &str = "git-stack";
     let mut success = true;
     if state.rebase {
         if state.repo.is_dirty() {
             return Err(proc_exit::Code::USAGE_ERR.with_message("Working tree is dirty, aborting"));
         }
 
-        let head_oid = state.repo.head_commit().id;
-        let head_branch = if let Some(branches) = state.branches.get(head_oid) {
-            branches[0].clone()
-        } else {
-            return Err(eyre::eyre!("Must not be in a detached HEAD state."))
-                .with_code(proc_exit::Code::USAGE_ERR);
-        };
+        let mut backups = git_stack::backup::Stack::new(BACKUP_NAME, &state.repo);
+        backups.capacity(state.backup_capacity);
+        let mut backup = git_stack::backup::Backup::from_repo(&state.repo)
+            .with_code(proc_exit::Code::FAILURE)?;
+        backup.insert_parent(&state.repo, &state.branches, &state.protected_branches);
+        backups.push(backup)?;
+
+        let head_branch = state
+            .repo
+            .head_branch()
+            .ok_or_else(|| eyre::eyre!("Must not be in a detached HEAD state."))
+            .with_code(proc_exit::Code::USAGE_ERR)?;
 
         let scripts: Result<Vec<_>, proc_exit::Exit> = state
             .stacks
@@ -270,6 +304,10 @@ pub fn stack(args: &crate::args::Args, colored_stdout: bool) -> proc_exit::ExitR
 
     show(&state, colored_stdout).with_code(proc_exit::Code::FAILURE)?;
 
+    if state.rebase {
+        log::info!("To undo, run `git branch-backup pop {}", BACKUP_NAME);
+    }
+
     if !success {
         return proc_exit::Code::FAILURE.ok();
     }
@@ -347,23 +385,19 @@ fn show(state: &State, colored_stdout: bool) -> eyre::Result<()> {
 
     match state.show_format {
         git_stack::config::Format::Silent => (),
-        git_stack::config::Format::Brief => {
+        git_stack::config::Format::Branches
+        | git_stack::config::Format::BranchCommits
+        | git_stack::config::Format::Commits => {
             writeln!(
                 std::io::stdout(),
                 "{}",
                 DisplayTree::new(&state.repo, &root)
                     .colored(colored_stdout)
-                    .all(false)
+                    .show(state.show_format)
             )?;
         }
-        git_stack::config::Format::Full => {
-            writeln!(
-                std::io::stdout(),
-                "{}",
-                DisplayTree::new(&state.repo, &root)
-                    .colored(colored_stdout)
-                    .all(true)
-            )?;
+        git_stack::config::Format::Debug => {
+            writeln!(std::io::stdout(), "{:#?}", root)?;
         }
     }
 
@@ -408,15 +442,17 @@ fn git_pull(
     repo: &mut git_stack::git::GitRepo,
     branch_name: &str,
     dry_run: bool,
-) -> eyre::Result<git2::Oid> {
+) -> eyre::Result<(git2::Oid, git2::Oid)> {
     let remote = repo.pull_remote();
     log::debug!("git pull --rebase {} {}", remote, branch_name);
     let remote_branch_name = format!("{}/{}", remote, branch_name);
     if dry_run {
-        return Ok(repo.find_local_branch(branch_name).unwrap().id);
+        let branch_id = repo.find_local_branch(branch_name).unwrap().id;
+        return Ok((branch_id, branch_id));
     }
 
-    let mut last_id;
+    let pulled_range;
+    let mut tip_id;
     {
         // A little uncertain about some of the weirder authentication needs, just deferring to `git`
         // instead of using `libgit2`
@@ -458,7 +494,8 @@ fn git_pull(
             remote_branch_name,
             remote_branch_annotated.id()
         );
-        last_id = remote_branch_annotated.id();
+        let end_id = remote_branch_annotated.id();
+        tip_id = end_id;
 
         let base_id = repo
             .merge_base(local_branch_annotated.id(), remote_branch_annotated.id())
@@ -472,12 +509,14 @@ fn git_pull(
         let base_annotated = repo.raw().find_annotated_commit(base_id).unwrap();
         log::trace!("rebase base {}", base_annotated.id());
 
-        if repo.merge_base(local_branch_annotated.id(), remote_branch_annotated.id())
-            == Some(remote_branch_annotated.id())
-        {
+        let merge_base_id =
+            repo.merge_base(local_branch_annotated.id(), remote_branch_annotated.id());
+        if merge_base_id == Some(remote_branch_annotated.id()) {
             log::debug!("{} is up-to-date with {}", branch_name, remote_branch_name);
-            return Ok(local_branch_annotated.id());
+            return Ok((local_branch_annotated.id(), local_branch_annotated.id()));
         }
+        let start_id = merge_base_id.unwrap_or(end_id);
+        pulled_range = (start_id, end_id);
 
         let mut rebase = repo
             .raw()
@@ -531,7 +570,7 @@ fn git_pull(
                         remote_branch_name
                     )
                 })?;
-            last_id = commit_id;
+            tip_id = commit_id;
         }
 
         rebase.finish(None).wrap_err_with(|| {
@@ -543,8 +582,9 @@ fn git_pull(
         })?;
     }
 
-    let local_branch = repo.find_local_branch(branch_name).unwrap();
-    if local_branch.id == repo.head_commit().id {
+    let head_branch = repo.head_branch();
+    let head_branch_name = head_branch.as_ref().map(|b| b.name.as_str());
+    if head_branch_name == Some(branch_name) {
         log::trace!("Updating {} (HEAD)", branch_name);
         repo.detach().wrap_err_with(|| {
             eyre::eyre!(
@@ -553,7 +593,7 @@ fn git_pull(
                 remote_branch_name
             )
         })?;
-        repo.branch(branch_name, last_id).wrap_err_with(|| {
+        repo.branch(branch_name, tip_id).wrap_err_with(|| {
             eyre::eyre!(
                 "failed to update `{}` to `{}`",
                 branch_name,
@@ -569,7 +609,7 @@ fn git_pull(
         })?;
     } else {
         log::trace!("Updating {}", branch_name);
-        repo.branch(branch_name, last_id).wrap_err_with(|| {
+        repo.branch(branch_name, tip_id).wrap_err_with(|| {
             eyre::eyre!(
                 "failed to update `{}` to `{}`",
                 branch_name,
@@ -578,7 +618,53 @@ fn git_pull(
         })?;
     }
 
-    Ok(last_id)
+    Ok(pulled_range)
+}
+
+fn drop_branches(
+    repo: &mut git_stack::git::GitRepo,
+    commit_ids: impl Iterator<Item = git2::Oid>,
+    potential_head: &str,
+    branches: &git_stack::git::Branches,
+    protected_branches: &git_stack::git::Branches,
+    dry_run: bool,
+) -> eyre::Result<()> {
+    let head_branch = repo.head_branch();
+    let head_branch_name = head_branch.as_ref().map(|b| b.name.as_str());
+
+    for commit_id in commit_ids {
+        let commit_branches: HashSet<_> = branches.get(commit_id).into_iter().flatten().collect();
+        let commit_protected_branches: HashSet<_> = protected_branches
+            .get(commit_id)
+            .into_iter()
+            .flatten()
+            .collect();
+        let mut commit_unprotected: Vec<_> = commit_branches
+            .difference(&commit_protected_branches)
+            .collect();
+        commit_unprotected.sort_unstable();
+        for branch in commit_unprotected {
+            if branch.name == potential_head {
+                continue;
+            } else if head_branch_name == Some(branch.name.as_str()) {
+                // Dom't leave HEAD detached but instead switch to the branch we pulled
+                log::trace!("git switch {}", potential_head);
+                if !dry_run {
+                    repo.switch(potential_head)?;
+                }
+                log::trace!("git branch -D {}", branch.name);
+                if !dry_run {
+                    repo.delete_branch(&branch.name)?;
+                }
+            } else {
+                log::trace!("git branch -D {}", branch.name);
+                if !dry_run {
+                    repo.delete_branch(&branch.name)?;
+                }
+            }
+        }
+    }
+    Ok(())
 }
 
 fn git_push(
@@ -650,7 +736,7 @@ struct DisplayTree<'r, 'n> {
     repo: &'r git_stack::git::GitRepo,
     root: &'n git_stack::graph::Node,
     palette: Palette,
-    all: bool,
+    show: git_stack::config::Format,
 }
 
 impl<'r, 'n> DisplayTree<'r, 'n> {
@@ -659,7 +745,7 @@ impl<'r, 'n> DisplayTree<'r, 'n> {
             repo,
             root,
             palette: Palette::plain(),
-            all: false,
+            show: Default::default(),
         }
     }
 
@@ -672,8 +758,8 @@ impl<'r, 'n> DisplayTree<'r, 'n> {
         self
     }
 
-    pub fn all(mut self, yes: bool) -> Self {
-        self.all = yes;
+    pub fn show(mut self, show: git_stack::config::Format) -> Self {
+        self.show = show;
         self
     }
 }
@@ -690,7 +776,7 @@ impl<'r, 'n> std::fmt::Display for DisplayTree<'r, 'n> {
             self.root.stacks.as_slice(),
             &mut tree,
             &self.palette,
-            self.all,
+            self.show,
         );
         tree.fmt(f)
     }
@@ -701,7 +787,7 @@ fn to_tree<'r, 'n, 'p>(
     nodes: &'n [Vec<git_stack::graph::Node>],
     tree: &mut treeline::Tree<RenderNode<'r, 'n, 'p>>,
     palette: &'p Palette,
-    show_all: bool,
+    show: git_stack::config::Format,
 ) {
     for branch in nodes {
         let mut branch_root = treeline::Tree::root(RenderNode {
@@ -710,7 +796,20 @@ fn to_tree<'r, 'n, 'p>(
             palette,
         });
         for node in branch {
-            if node.branches.is_empty() && node.stacks.is_empty() && !show_all {
+            let skip = match show {
+                git_stack::config::Format::Silent => true,
+                git_stack::config::Format::Commits => false,
+                git_stack::config::Format::BranchCommits => {
+                    let protected = node.action.is_protected() || node.action.is_rebase();
+                    protected && node.branches.is_empty()
+                }
+                git_stack::config::Format::Branches => {
+                    let boring_commit = node.branches.is_empty() && node.stacks.is_empty();
+                    boring_commit
+                }
+                git_stack::config::Format::Debug => false,
+            };
+            if skip {
                 log::trace!("Skipping commit {}", node.local_commit.id);
                 continue;
             }
@@ -719,13 +818,7 @@ fn to_tree<'r, 'n, 'p>(
                 node: Some(node),
                 palette,
             });
-            to_tree(
-                repo,
-                node.stacks.as_slice(),
-                &mut child_tree,
-                palette,
-                show_all,
-            );
+            to_tree(repo, node.stacks.as_slice(), &mut child_tree, palette, show);
             branch_root.push(child_tree);
         }
         tree.push(branch_root);
@@ -745,111 +838,42 @@ impl<'r, 'n, 'p> std::fmt::Display for RenderNode<'r, 'n, 'p> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
         if let Some(node) = self.node.as_ref() {
             if node.branches.is_empty() {
-                if node.stacks.is_empty() {
-                    write!(f, "{} ", self.palette.info.paint(node.local_commit.id))?;
+                let abbrev_id = self
+                    .repo
+                    .raw()
+                    .find_object(node.local_commit.id, None)
+                    .unwrap()
+                    .short_id()
+                    .unwrap();
+                let style = if node.stacks.is_empty() {
+                    self.palette.info
                 } else {
                     // Branches should be off of other branches
-                    write!(f, "{} ", self.palette.warn.paint(node.local_commit.id))?;
-                }
-            } else if node.action.is_protected() || node.action.is_rebase() {
-                write!(
-                    f,
-                    "{} ",
-                    self.palette
-                        .info
-                        .paint(node.branches.iter().map(|b| b.name.as_str()).join(", ")),
-                )?;
+                    self.palette.warn
+                };
+                write!(f, "{}", style.paint(abbrev_id.as_str().unwrap()))?;
             } else {
                 write!(
                     f,
-                    "{} ",
-                    self.palette
-                        .branch
-                        .paint(node.branches.iter().map(|b| b.name.as_str()).join(", ")),
+                    "{}",
+                    node.branches
+                        .iter()
+                        .map(|b| {
+                            format!(
+                                "{}{}",
+                                format_branch_name(b, node, self.palette),
+                                format_branch_status(b, self.repo, node, self.palette),
+                            )
+                        })
+                        .join(", ")
                 )?;
             }
 
-            if node.action.is_protected() || node.action.is_rebase() {
-                if !node.branches.is_empty() {
-                    let branch = &node.branches[0];
-                    match commit_relation(self.repo, branch.id, branch.pull_id) {
-                        Some((0, 0)) => {}
-                        Some((local, 0)) => {
-                            write!(
-                                f,
-                                "{} ",
-                                self.palette.warn.paint(format!("({} ahead)", local)),
-                            )?;
-                        }
-                        Some((0, remote)) => {
-                            write!(
-                                f,
-                                "{} ",
-                                self.palette.warn.paint(format!("({} behind)", remote)),
-                            )?;
-                        }
-                        Some((local, remote)) => {
-                            write!(
-                                f,
-                                "{} ",
-                                self.palette
-                                    .warn
-                                    .paint(format!("({} ahead, {} behind)", local, remote)),
-                            )?;
-                        }
-                        None => {
-                            write!(f, "{} ", self.palette.warn.paint("(no remote)"))?;
-                        }
-                    }
-                }
-            } else if 1 < self
-                .repo
-                .raw()
-                .find_commit(node.local_commit.id)
-                .unwrap()
-                .parent_count()
-            {
-                write!(f, "{} ", self.palette.error.paint("merge commit"))?;
-            } else {
-                if !node.branches.is_empty() {
-                    if node.pushable {
-                        write!(f, "{} ", self.palette.info.paint("(ready) "))?;
-                    } else {
-                        let branch = &node.branches[0];
-                        match commit_relation(self.repo, branch.id, branch.push_id) {
-                            Some((0, 0)) => {
-                                write!(f, "{} ", self.palette.info.paint("(pushed)"))?;
-                            }
-                            Some((local, 0)) => {
-                                write!(
-                                    f,
-                                    "{} ",
-                                    self.palette.info.paint(format!("({} ahead)", local)),
-                                )?;
-                            }
-                            Some((0, remote)) => {
-                                write!(
-                                    f,
-                                    "{} ",
-                                    self.palette.warn.paint(format!("({} behind)", remote)),
-                                )?;
-                            }
-                            Some((local, remote)) => {
-                                write!(
-                                    f,
-                                    "{} ",
-                                    self.palette
-                                        .warn
-                                        .paint(format!("({} ahead, {} behind)", local, remote)),
-                                )?;
-                            }
-                            None => {
-                                write!(f, "{} ", self.palette.info.paint("(no remote)"))?;
-                            }
-                        }
-                    }
-                }
-            }
+            write!(
+                f,
+                "{} ",
+                format_commit_status(self.repo, node, self.palette)
+            )?;
 
             let summary = String::from_utf8_lossy(&node.local_commit.summary);
             if node.action.is_protected() || node.action.is_rebase() {
@@ -870,6 +894,112 @@ impl<'r, 'n, 'p> std::fmt::Display for RenderNode<'r, 'n, 'p> {
     }
 }
 
+fn format_branch_name<'d>(
+    branch: &'d git_stack::git::Branch,
+    node: &'d git_stack::graph::Node,
+    palette: &'d Palette,
+) -> impl std::fmt::Display + 'd {
+    if node.action.is_protected() || node.action.is_rebase() {
+        palette.info.paint(branch.name.as_str())
+    } else {
+        palette.good.paint(branch.name.as_str())
+    }
+}
+
+fn format_branch_status<'d>(
+    branch: &'d git_stack::git::Branch,
+    repo: &'d git_stack::git::GitRepo,
+    node: &'d git_stack::graph::Node,
+    palette: &'d Palette,
+) -> String {
+    // See format_commit_status
+    if node.action.is_protected() || node.action.is_rebase() {
+        match commit_relation(repo, branch.id, branch.pull_id) {
+            Some((0, 0)) => format!(""),
+            Some((local, 0)) => {
+                format!(" {}", palette.warn.paint(format!("({} ahead)", local)))
+            }
+            Some((0, remote)) => {
+                format!(" {}", palette.warn.paint(format!("({} behind)", remote)))
+            }
+            Some((local, remote)) => {
+                format!(
+                    " {}",
+                    palette
+                        .warn
+                        .paint(format!("({} ahead, {} behind)", local, remote)),
+                )
+            }
+            None => {
+                format!(" {}", palette.warn.paint("(no remote)"))
+            }
+        }
+    } else if node.action.is_delete() {
+        format!("")
+    } else if 1 < repo
+        .raw()
+        .find_commit(node.local_commit.id)
+        .unwrap()
+        .parent_count()
+    {
+        format!("")
+    } else {
+        if node.branches.is_empty() {
+            format!("")
+        } else {
+            if node.pushable {
+                format!(" {}", palette.info.paint("(ready) "))
+            } else {
+                let branch = &node.branches[0];
+                match commit_relation(repo, branch.id, branch.push_id) {
+                    Some((0, 0)) => {
+                        format!(" {}", palette.good.paint("(pushed)"))
+                    }
+                    Some((local, 0)) => {
+                        format!(" {}", palette.info.paint(format!("({} ahead)", local)))
+                    }
+                    Some((0, remote)) => {
+                        format!(" {}", palette.warn.paint(format!("({} behind)", remote)))
+                    }
+                    Some((local, remote)) => {
+                        format!(
+                            " {}",
+                            palette
+                                .warn
+                                .paint(format!("({} ahead, {} behind)", local, remote)),
+                        )
+                    }
+                    None => {
+                        format!(" {}", palette.info.paint("(no remote)"))
+                    }
+                }
+            }
+        }
+    }
+}
+
+fn format_commit_status<'d>(
+    repo: &'d git_stack::git::GitRepo,
+    node: &'d git_stack::graph::Node,
+    palette: &'d Palette,
+) -> String {
+    // See format_commit_status
+    if node.action.is_protected() || node.action.is_rebase() {
+        format!("")
+    } else if node.action.is_delete() {
+        format!("{} ", palette.warn.paint("(drop)"))
+    } else if 1 < repo
+        .raw()
+        .find_commit(node.local_commit.id)
+        .unwrap()
+        .parent_count()
+    {
+        format!("{} ", palette.error.paint("(merge commit)"))
+    } else {
+        format!("")
+    }
+}
+
 fn commit_relation(
     repo: &git_stack::git::GitRepo,
     local: git2::Oid,
@@ -893,7 +1023,7 @@ struct Palette {
     error: yansi::Style,
     warn: yansi::Style,
     info: yansi::Style,
-    branch: yansi::Style,
+    good: yansi::Style,
     hint: yansi::Style,
 }
 
@@ -903,7 +1033,7 @@ impl Palette {
             error: yansi::Style::new(yansi::Color::Red),
             warn: yansi::Style::new(yansi::Color::Yellow),
             info: yansi::Style::new(yansi::Color::Blue),
-            branch: yansi::Style::new(yansi::Color::Green),
+            good: yansi::Style::new(yansi::Color::Green),
             hint: yansi::Style::new(yansi::Color::Blue).dimmed(),
         }
     }
@@ -913,7 +1043,7 @@ impl Palette {
             error: yansi::Style::default(),
             warn: yansi::Style::default(),
             info: yansi::Style::default(),
-            branch: yansi::Style::default(),
+            good: yansi::Style::default(),
             hint: yansi::Style::default(),
         }
     }
