@@ -7,6 +7,7 @@ pub trait Repo {
 
     fn find_commit(&self, id: git2::Oid) -> Option<std::rc::Rc<Commit>>;
     fn head_commit(&self) -> std::rc::Rc<Commit>;
+    fn head_branch(&self) -> Option<Branch>;
     fn resolve(&self, revspec: &str) -> Option<std::rc::Rc<Commit>>;
     fn commits_from(
         &self,
@@ -19,6 +20,7 @@ pub trait Repo {
     ) -> Result<git2::Oid, git2::Error>;
 
     fn branch(&mut self, name: &str, id: git2::Oid) -> Result<(), git2::Error>;
+    fn delete_branch(&mut self, name: &str) -> Result<(), git2::Error>;
     fn find_local_branch(&self, name: &str) -> Option<Branch>;
     fn local_branches(&self) -> Box<dyn Iterator<Item = Branch> + '_>;
     fn detach(&mut self) -> Result<(), git2::Error>;
@@ -163,6 +165,36 @@ impl GitRepo {
         self.find_commit(head_id).unwrap()
     }
 
+    pub fn head_branch(&self) -> Option<Branch> {
+        let resolved = self.repo.head().unwrap().resolve().unwrap();
+        let name = resolved.shorthand()?;
+        let id = resolved.target()?;
+
+        let push_id = self
+            .repo
+            .find_branch(
+                &format!("{}/{}", self.push_remote(), name),
+                git2::BranchType::Remote,
+            )
+            .ok()
+            .and_then(|b| b.get().target());
+        let pull_id = self
+            .repo
+            .find_branch(
+                &format!("{}/{}", self.pull_remote(), name),
+                git2::BranchType::Remote,
+            )
+            .ok()
+            .and_then(|b| b.get().target());
+
+        Some(Branch {
+            name: name.to_owned(),
+            id,
+            push_id,
+            pull_id,
+        })
+    }
+
     pub fn resolve(&self, revspec: &str) -> Option<std::rc::Rc<Commit>> {
         let id = self.repo.revparse_single(revspec).ok()?.id();
         self.find_commit(id)
@@ -185,51 +217,73 @@ impl GitRepo {
         head_id: git2::Oid,
         cherry_id: git2::Oid,
     ) -> Result<git2::Oid, git2::Error> {
-        // Based on https://www.pygit2.org/recipes/git-cherry-pick.html
-        let base_id = self.repo.merge_base(head_id, cherry_id)?;
-        let base_commit = self.repo.find_commit(base_id)?;
-        let base_tree = self.repo.find_tree(base_commit.tree_id())?;
-
-        let head_commit = self.repo.find_commit(head_id)?;
-        let head_tree = self.repo.find_tree(head_commit.tree_id())?;
-
-        let cherry_commit = self.repo.find_commit(cherry_id)?;
-        let cherry_tree = self.repo.find_tree(cherry_commit.tree_id())?;
-
-        let mut result_index = self
-            .repo
-            .merge_trees(&base_tree, &head_tree, &cherry_tree, None)?;
-        if result_index.has_conflicts() {
-            let conflicts = result_index
-                .conflicts()?
-                .map(|conflict| {
-                    let conflict = conflict.unwrap();
-                    let our_path = conflict
-                        .our
-                        .as_ref()
-                        .map(|c| bytes2path(&c.path))
-                        .or_else(|| conflict.their.as_ref().map(|c| bytes2path(&c.path)))
-                        .unwrap();
-                    format!("{}", our_path.display())
-                })
-                .join("\n  ");
-            return Err(git2::Error::new(
-                git2::ErrorCode::Unmerged,
-                git2::ErrorClass::Index,
-                format!("cherry-pick conflicts:\n  {}\n", conflicts),
-            ));
+        let base_id = self
+            .commits_from(cherry_id)
+            .filter(|c| c.id != cherry_id)
+            .map(|c| c.id)
+            .next()
+            .unwrap_or(cherry_id);
+        if base_id == head_id {
+            return Ok(cherry_id);
         }
-        let result_id = result_index.write_tree_to(&self.repo)?;
-        let result_tree = self.repo.find_tree(result_id)?;
-        let new_id = self.repo.commit(
-            None,
-            &cherry_commit.author(),
-            &cherry_commit.committer(),
-            cherry_commit.message().unwrap(),
-            &result_tree,
-            &[&head_commit],
+        let base_commit = self.repo.find_annotated_commit(base_id)?;
+        let head_commit = self.repo.find_annotated_commit(head_id)?;
+        let cherry_commit = self.repo.find_annotated_commit(cherry_id)?;
+        let mut rebase = self.repo.rebase(
+            Some(&cherry_commit),
+            Some(&base_commit),
+            Some(&head_commit),
+            Some(git2::RebaseOptions::new().inmemory(true)),
         )?;
-        Ok(new_id)
+
+        let mut tip_id = head_id;
+        while let Some(op) = rebase.next() {
+            op.map_err(|e| {
+                let _ = rebase.abort();
+                e
+            })?;
+            let inmemory_index = rebase.inmemory_index().unwrap();
+            if inmemory_index.has_conflicts() {
+                let conflicts = inmemory_index
+                    .conflicts()?
+                    .map(|conflict| {
+                        let conflict = conflict.unwrap();
+                        let our_path = conflict
+                            .our
+                            .as_ref()
+                            .map(|c| bytes2path(&c.path))
+                            .or_else(|| conflict.their.as_ref().map(|c| bytes2path(&c.path)))
+                            .unwrap();
+                        format!("{}", our_path.display())
+                    })
+                    .join("\n  ");
+                return Err(git2::Error::new(
+                    git2::ErrorCode::Unmerged,
+                    git2::ErrorClass::Index,
+                    format!("cherry-pick conflicts:\n  {}\n", conflicts),
+                ));
+            }
+
+            let sig = self.repo.signature().unwrap();
+            let commit_id = match rebase.commit(None, &sig, None).map_err(|e| {
+                let _ = rebase.abort();
+                e
+            }) {
+                Ok(commit_id) => Ok(commit_id),
+                Err(err) => {
+                    if err.class() == git2::ErrorClass::Rebase
+                        && err.code() == git2::ErrorCode::Applied
+                    {
+                        log::trace!("Skipping {}, already applied to {}", cherry_id, head_id);
+                        return Ok(tip_id);
+                    }
+                    Err(err)
+                }
+            }?;
+            tip_id = commit_id;
+        }
+        rebase.finish(None)?;
+        Ok(tip_id)
     }
 
     pub fn branch(&mut self, name: &str, id: git2::Oid) -> Result<(), git2::Error> {
@@ -238,6 +292,12 @@ impl GitRepo {
         Ok(())
     }
 
+    pub fn delete_branch(&mut self, name: &str) -> Result<(), git2::Error> {
+        // HACK: We shouldn't limit ourselves to `Local`
+        let mut branch = self.repo.find_branch(name, git2::BranchType::Local)?;
+        branch.delete()
+    }
+
     pub fn find_local_branch(&self, name: &str) -> Option<Branch> {
         let branch = self.repo.find_branch(name, git2::BranchType::Local).ok()?;
         let id = branch.get().target().unwrap();
@@ -326,6 +386,7 @@ impl GitRepo {
     }
 
     pub fn switch(&mut self, name: &str) -> Result<(), git2::Error> {
+        // HACK: We shouldn't limit ourselves to `Local`
         let branch = self.repo.find_branch(name, git2::BranchType::Local)?;
         self.repo.set_head(branch.get().name().unwrap())?;
         let mut builder = git2::build::CheckoutBuilder::new();
@@ -352,6 +413,10 @@ impl Repo for GitRepo {
         self.head_commit()
     }
 
+    fn head_branch(&self) -> Option<Branch> {
+        self.head_branch()
+    }
+
     fn resolve(&self, revspec: &str) -> Option<std::rc::Rc<Commit>> {
         self.resolve(revspec)
     }
@@ -375,6 +440,10 @@ impl Repo for GitRepo {
         self.branch(name, id)
     }
 
+    fn delete_branch(&mut self, name: &str) -> Result<(), git2::Error> {
+        self.delete_branch(name)
+    }
+
     fn find_local_branch(&self, name: &str) -> Option<Branch> {
         self.find_local_branch(name)
     }
@@ -465,6 +534,13 @@ impl InMemoryRepo {
         self.commits.get(&self.head_id.unwrap()).cloned().unwrap().1
     }
 
+    pub fn head_branch(&self) -> Option<Branch> {
+        self.branches
+            .values()
+            .find(|b| b.id == self.head_id.unwrap())
+            .cloned()
+    }
+
     pub fn resolve(&self, revspec: &str) -> Option<std::rc::Rc<Commit>> {
         let branch = self.branches.get(revspec)?;
         self.find_commit(branch.id)
@@ -514,6 +590,16 @@ impl InMemoryRepo {
         Ok(())
     }
 
+    pub fn delete_branch(&mut self, name: &str) -> Result<(), git2::Error> {
+        self.branches.remove(name).map(|_| ()).ok_or_else(|| {
+            git2::Error::new(
+                git2::ErrorCode::NotFound,
+                git2::ErrorClass::Reference,
+                format!("could not remove branch {:?}", name),
+            )
+        })
+    }
+
     pub fn find_local_branch(&self, name: &str) -> Option<Branch> {
         self.branches.get(name).cloned()
     }
@@ -600,10 +686,18 @@ impl Repo for InMemoryRepo {
         self.cherry_pick(head_id, cherry_id)
     }
 
+    fn head_branch(&self) -> Option<Branch> {
+        self.head_branch()
+    }
+
     fn branch(&mut self, name: &str, id: git2::Oid) -> Result<(), git2::Error> {
         self.branch(name, id)
     }
 
+    fn delete_branch(&mut self, name: &str) -> Result<(), git2::Error> {
+        self.delete_branch(name)
+    }
+
     fn find_local_branch(&self, name: &str) -> Option<Branch> {
         self.find_local_branch(name)
     }
