@@ -3,6 +3,7 @@ pub enum Action {
     Pick,
     Protected,
     Rebase(git2::Oid),
+    Delete,
 }
 
 impl Action {
@@ -17,4 +18,8 @@ impl Action {
     pub fn is_rebase(&self) -> bool {
         matches!(self, Action::Rebase(_))
     }
+
+    pub fn is_delete(&self) -> bool {
+        matches!(self, Action::Delete)
+    }
 }
