@@ -1,6 +1,9 @@
+#![allow(clippy::collapsible_else_if)]
+
 #[macro_use]
 extern crate clap;
 
+pub mod backup;
 pub mod color;
 pub mod config;
 pub mod git;
